/* -*- mode: c++; indent-tabs-mode:nil -*- */
#include <iostream>
#include <iomanip>

#include <util.h>
#include <jengine.h>
#include <jplus.h>
#include <yacts.h>

using namespace std;

// Testsuite
int tests_run=0;
int tests_ok=0;
int tests_fail=0;

const int test_field_width=75;
#define starttest(message) do { stringstream ss; ss << message; cout << left << setw(test_field_width)<< ss.str().substr(0,test_field_width) << ':'; tests_run++; } while (0)
#define section(message) do { cout << endl << "=== " << message << endl;} while(0)
#define testassert(test) do { if (test) {cout<<"ok"<<endl; tests_ok++;} else {cout<<"FAIL"<<endl; tests_fail++;} } while (0)
const lReal tests_eq_rel=0.0002;
const lReal tests_eq_abs=1e-6;
bool tol_equ(Real a, Real b) {
  return (abs(a-b)<tests_eq_abs) || (abs(a-b)<=tests_eq_rel*max(abs(a),abs(b)));
};
#define testcompare(a,b) (tol_equ(a,b))
#define testfloateq(a,b) do {bool r=testcompare((a),(b)); testassert(r); if (!r) cout << "Got " << (a) << " expected " << (b) << "." << endl; } while (0)
#define testeq(a,b) do {bool r=((a)==(b)); testassert(r); if (!r) cout << "Got \"" << (a) << "\" expected \"" << (b) << "\"." << endl; } while (0)
#define testJbuf(b) do {string a=jout.str(), bb=b, e=bb+(bb.length()>0?"\n":""); jout.str(""); testeq(a,e); } while (0)
#define testJ(s,b) do { starttest("doJ: "<<s); JE.doJ(s);  testJbuf(b); } while (0)
#define testY(s,b) do { starttest("doJ: "<<s); jout.clear(); YACTS.doJ(s);  testJbuf(b); } while (0)

jarray mf1(jarray y) {
  // here we simply return J array back
  return y;
};

jarray df1(jarray x, jarray y) {
  // here we simply return J array back
  return y;
};

jarray mf2(jarray y) {
  // here we allocate our own
  jengine *je=y.getEngine();
  vector<jarray::I> shape;
  shape.push_back(y.size());
  jarray r(je, jarray::T_FL,shape);
  for(int i=0; i<y.size(); i++) {
    double v;
    y.get(v,i);
    r.set(i,v+1.0);
  };
  return r;
};

jarray df2(jarray x, jarray y) {
  jengine *je=y.getEngine();
  double q=1.0;
  if (x.size()>0) x.get(q,0);
  // here we allocate our own
  vector<jarray::I> shape;
  shape.push_back(y.size());
  jarray r(je,jarray::T_FL,shape);
  for(int i=0; i<y.size(); i++) {
    double v;
    y.get(v,i);
    r.set(i,v+q);
  };
  return r;
};

jarray amf1(jengine::monad um, jengine::dyad ud, jarray y) {
  // here we simply return J array back
  return y;
};

jarray adf1(jengine::monad um, jengine::dyad ud, jarray x, jarray y) {
  // here we simply return J array back
  return y;
};

jarray amf2(jengine::monad um, jengine::dyad ud, jarray y) {
  // we apply verb argument to y monadically
  return um(y);
};

jarray adf2(jengine::monad um, jengine::dyad ud, jarray x, jarray y) {
  // we apply verb argument to x and y dyadically
  return ud(x,y);
};

jarray mf3(jarray y) {
  jengine *je=y.getEngine();
  jarray_of_type<double> yd(y);
  jarray_of_type<double> rv(je,y.size());
  for(int i=0;i<y.size();i++) rv(i)=yd(i)+1.0;
  return rv;
};

// function, which should receive at most vectors and returns the rank
jarray mfr(jarray y) {
  jengine *je=y.getEngine();
  jarray_of_type<jarray::I> rv(je);
  rv[0]=y.rank();
  return rv;
};


// this verb sets the flag when it was called, used to check memoization.
bool monadcalled;
bool dyadcalled;
jarray vmark(jarray y) {
  monadcalled=true;
  return y;
};
jarray vmark(jarray x,jarray y) {
  dyadcalled=true;
  return y;
};


int runTests() {
  section("testsuite tests");
  starttest("Basic testsuite test");
  testassert(true);
  starttest("Basic testsuite output test (passed so far "<<tests_ok<<")");
  testassert(tests_ok==1);
  starttest("Basic testsuite too long output test"<<setfill('_')<<setw(80)<<"");
  testassert(true);
  starttest("Basic testsuite failing test (this fail does not count)");
  testassert(false);
  tests_fail--; tests_ok++;
  {
    Real a = 0.1;
    Real b = 0.100001;
    starttest("Basic testsuite test that "<<a<<"=="<<b<<" (should be)");
    testfloateq(a,b);
  }
  {
    Real a = 0.1;
    Real b = 0.1001;
    starttest("Basic testsuite test that "<<a<<"=="<<b<<" (should fail, its ok)");
    testfloateq(a,b);
    if (!testcompare(a,b)) {
      tests_fail--; tests_ok++;
    } else { // if test did not, actually, fail -- it's a failure
      tests_fail++; tests_ok--;
    };
  }
  stringstream jout;
  {
    starttest("Load and initialize J library");
    jengine::initJlibrary(jout);
    testassert(true);
  }
  {
    jengine JE;


    section("testing simple scripts and variable access");
    testJ("'Hello World!'","Hello World!");
    testJ("a=:2 1 3","");
    testJ("$ a","3");
    {
      starttest("JE.get(\"a\")");
      jarray a=JE.get("a");
      testassert(a.isValid());
      starttest("a.shape(...)");
      vector<jarray::I> shape;
      a.shape(shape);
      testassert(true);
      starttest("a is vector");
      testeq(shape.size(),1);
      starttest("a has length 3");
      testeq(shape[0],3);
      starttest("stream << a = \"2 1 3\"");    
      stringstream ss;
      ss << a;
      testeq(ss.str(),"2 1 3");
    }
    {
      starttest("allocate 4 element double vector");
      vector <jarray::I> shape;
      shape.push_back(4);
      //jarray a(&JE,jarray::T_FL, shape);
      jarray a(NULL,jarray::T_FL, shape);
      testassert(a.isValid());
      starttest("initialize its elements to {3,4,5,6}");
      a.set(0,3);
      a.set(1,4);
      a.set(2,5);
      a.set(3,6);
      testassert(a.isValid());
      starttest("stream << a = \"3 4 5 6\"");    
      stringstream ss;
      ss << a;
      testeq(ss.str(),"3 4 5 6");
      starttest("pass to J as \"a\"");    
      testassert(JE.set("a",a));
    }
    testJ(", a","3 4 5 6");
    
    section("testing verbs");
    {
      starttest("defining test verb v1 (monadic)");
      testassert(JE.defVerb("v1", mf1, NULL));
    }
    testJ("2 {. v1 10000 $ 1.1 1.2","1.1 1.2");
    {
      starttest("defining test verb v2 (monadic)");
      testassert(JE.defVerb("v2", mf2, NULL));
    }
    testJ("v2 1.1 1.2","2.1 2.2");
    testJ("c =: v2 1.1 1.2","");
    testJ("c","2.1 2.2");
    testJ("v1 c","2.1 2.2");
    testJ("v1 c","2.1 2.2");  
    testJ("v1 v1 c","2.1 2.2");  
    testJ("v1 v1 v1 c","2.1 2.2");
    
    testJ("v2 c","3.1 3.2");  
    testJ("c =: 2.1 2.2","");
    testJ("v2 c","3.1 3.2");  
    testJ("1.3 v2 1.1","|domain error: v2\n|   1.3     v2 1.1");
    {
      starttest("defining test verb v2 (dyadic only)");
      testassert(JE.defVerb("v2", NULL, df2));
    }
    testJ("1.1 v2 1.1 1.2", "2.2 2.3");
    testJ("v2 1.1 1.2","|domain error: v2\n|       v2 1.1 1.2");
    {
      starttest("defining test verb v1 (monadic and dyadic)");
      testassert(JE.defVerb("v1", mf1, df1));
    }
    testJ("1.1 v1 1.1 1.2","1.1 1.2");
    testJ("c =: 1 v2 1.1 1.2","");
    testJ("c","2.1 2.2");
    testJ("1&v1 c","2.1 2.2");
    testJ("1&v1 c","2.1 2.2");  
    testJ("1&v1 1&v1 c","2.1 2.2");  
    testJ("1&v1 1&v1 1&v1 c","2.1 2.2");
    {
      starttest("defining test verb v2 (monadic and dyadic)");
      testassert(JE.defVerb("v2", mf2, df2));
    }
    testJ("2.0 v2 1.1 1.2","3.1 3.2");
    testJ("c =: 2.0 v2 1.1 1.2","");
    testJ("c","3.1 3.2");  
    testJ("2.0 v2 c","5.1 5.2");
    testJ("c =: 3.1 3.2","");
    testJ("2.0 v2 c","5.1 5.2");  
    
    section("testing adverbs");
    {
      starttest("defining test adverb av1 (monadic)");
      testassert(JE.defAdverb("av1", amf1, NULL));
    }
    testJ("+ av1 1.1 1.2","1.1 1.2");
    testJ("+ av1 + av1 1.1 1.2","1.1 1.2");
    testJ("+ av1 + av1 + av1 1.1 1.2","1.1 1.2");
    {
      starttest("defining test adverb av1 (monadic and dyadic)");
      testassert(JE.defAdverb("av1", amf1, adf1));
    }
    testJ("1.2 (+ av1) 1.1 1.2","1.1 1.2");
    {
      starttest("defining test adverb av2 (monadic), based on mf1");
      testassert(JE.defAdverb("av2", amf1, NULL));
    }
    testJ("+ av2 1.1 1.2","1.1 1.2");
    
    {
      starttest("defining test adverb av2 (monadic and dyadic)");
      testassert(JE.defAdverb("av2", amf2, adf2));
    }
    testJ("c =: 1.1 1.2","");
    testJ("+: av2 c","2.2 2.4");
    testJ("1.2 (+ av2) c","2.3 2.4");

    section("different combinations of nested verbs ans adverbs");
    testJ("(v1 av2) c","1.1 1.2");
    testJ("(v2 av2) c","2.1 2.2");
    testJ("((v1 av2) av2) c","1.1 1.2");
    testJ("(((v1 av2) av2) av2) c","1.1 1.2");
    testJ("((v2 av2) av2) c","2.1 2.2");
    testJ("((2.0&(v2 av2)) av2) c","3.1 3.2");

    section("testing jarray_of_type");
    {
      starttest("defining test verb v3 (monadic), which uses jarray_of_type");
      testassert(JE.defVerb("v3", mf3, NULL));
    }
    testJ("v3 1.1 1.2","2.1 2.2");
    {
      starttest("constructing and passing 2x3 double array  \"a\" (2 rows, 3 cols)");
      jarray_of_type<double> a(&JE,2,3);
      a(0,0)=1;a(0,1)=2;a(0,2)=3;
      a(1,0)=4;a(1,1)=5;a(1,2)=6;
      testassert(JE.set("a",a));
      testJ("a","1 2 3\n4 5 6");
    }
    {
      starttest("constructing and passing 2x3x4 double array \"a\"");
      jarray_of_type<double> a(&JE,2,3,4);
      a(0,0,0)=1;a(0,0,1)=2;a(0,0,2)=3;a(0,0,3)=4;
      a(0,1,0)=5;a(0,1,1)=6;a(0,1,2)=7;a(0,1,3)=8;
      a(0,2,0)=9;a(0,2,1)=10;a(0,2,2)=11;a(0,2,3)=12;

      a(1,0,0)=13;a(1,0,1)=14;a(1,0,2)=15;a(1,0,3)=16;
      a(1,1,0)=17;a(1,1,1)=18;a(1,1,2)=19;a(1,1,3)=20;
      a(1,2,0)=21;a(1,2,1)=22;a(1,2,2)=23;a(1,2,3)=24;
      testassert(JE.set("a",a));
      testJ("a"," 1  2  3  4\n 5  6  7  8\n 9 10 11 12\n"
            "\n"
            "13 14 15 16\n17 18 19 20\n21 22 23 24");
    }
    {
      starttest("constructing and passing 2x3x4 double array \"a\" (generic)");
      vector<jarray::I> shape(3);
      shape[0]=2; shape[1]=3; shape[2]=4;
      jarray_of_type<double> a(&JE,shape);
      vector<jarray::I> loc(3);
      loc[0]=0; loc[1]=0; loc[2]=0;
      a(loc)=1; loc[2]=1; a(loc)=2; loc[2]=2; a(loc)=3; loc[2]=3; a(loc)=4;
      loc[1]=1; loc[2]=0;
      a(loc)=5; loc[2]=1; a(loc)=6; loc[2]=2; a(loc)=7; loc[2]=3; a(loc)=8;
      loc[1]=2; loc[2]=0;
      a(loc)=9; loc[2]=1; a(loc)=10; loc[2]=2; a(loc)=11; loc[2]=3; a(loc)=12;

      loc[0]=1; loc[1]=0; loc[2]=0;
      a(loc)=13; loc[2]=1; a(loc)=14; loc[2]=2; a(loc)=15; loc[2]=3; a(loc)=16;
      loc[1]=1; loc[2]=0;
      a(loc)=17; loc[2]=1; a(loc)=18; loc[2]=2; a(loc)=19; loc[2]=3; a(loc)=20;
      loc[1]=2; loc[2]=0;
      a(loc)=21; loc[2]=1; a(loc)=22; loc[2]=2; a(loc)=23; loc[2]=3; a(loc)=24;
      testassert(JE.set("a",a));
      testJ("a"," 1  2  3  4\n 5  6  7  8\n 9 10 11 12\n"
            "\n"
            "13 14 15 16\n17 18 19 20\n21 22 23 24");
    }

    section("testing verb rank support");
    {
      starttest("defining test verb vr (monadic) with ranks 1");
      testassert(JE.defVerb("vr", mfr, NULL, 1, 1, 1));
    }
    testJ("vr 1.1 1.2","1");
    testJ("vr 1.1 1.2 ,. 1.3 1.4","1 1");
    testJ("vr 1.1 1.2 ,. 1.3 1.4 ,. 1.5 1.6 ","1 1");
    testJ("vr |: 1.1 1.2 ,. 1.3 1.4 ,. 1.5 1.6 ","1 1 1");
  }

  section("trivial J+ tests");
  {
    string script="   a=:1\n   b=:2\n   c=:a+b";
    cout<<"context is provided by the following J+ script:"<<endl<<script<<endl;
    istringstream in(script);
    starttest("Checking that script can be successfully parsed");
    jplus JP;
    JP.init(in);   
    testassert(JP.ok());
    starttest("Pulling \"a\"");
    jarray a = JP.get("a");
    testassert(JP.ok());
    starttest("Checking that \"a\" is indeed scalar boolean");
    testassert((a.size()==1)&&(a.rank()==0)&&(a.type()==jarray::T_B01));
    starttest("Checking that \"a\" is indeed 1");
    JP.doJ("a");
    jarray_of_type<jarray::I> ai(a);
    testeq(ai[0],1);
    starttest("Checking that \"b\" is indeed scalar integer");
    jarray b = JP.get("b");
    testassert((b.size()==1)&&(b.rank()==0)&&(b.type()==jarray::T_INT));
    starttest("Checking that \"b\" is indeed 2");
    jarray_of_type<jarray::I> bi(b);
    testeq(bi[0],2);
    starttest("Checking that \"c\" is indeed scalar integer");
    jarray c = JP.get("c");
    testassert((c.size()==1)&&(c.rank()==0)&&(c.type()==jarray::T_INT));
    starttest("Checking that \"c\" is indeed 3");
    jarray_of_type<jarray::I> ci(c);
    testeq(ci[0],3);
  }
  
  section("yacts tests");
  {
    starttest("Instantiate yacts");
    istringstream in("");
    yacts YACTS;
    YACTS.init(in);
    testassert(true);
    testY("1","1\n1"); // very strange an extra "1" is printed somewhere
    testY("fft 1 2 3 4","10 _2j2 _2 _2j_2");
    testY("(#@,@] %~ 1 fft fft) 1 2 3 4 ,: 5 6 7 8",
          "1 2 3 4\n5 6 7 8");
    testY("(#@,@] %~ $@] fftc2r fftr2c) 1 2 3 4 5 6 7",
          "1 2 3 4 5 6 7");
    testY("([: % *:) int 1 _","1");
    testY("([: +/ *:) int 1 3 , 2 3 ,: 4 5","62");
    {
      starttest("defining vmark verb to test memoization");
      testassert(YACTS.defVerb("vmark", vmark, vmark));
    }
    monadcalled=false;
    dyadcalled=false;
    testY("vmark 1","1");
    starttest("monad called");
    testassert(monadcalled);
    testY("1 vmark 1","1");
    starttest("dyad called");
    testassert(dyadcalled);
    testY("(vmark memo) 1","1");
    testY("1 (vmark memo) 1","1");
    monadcalled=false;
    dyadcalled=false;
    testY("(vmark memo) 1","1");
    starttest("cache hit, monad NOT called");
    testassert(!monadcalled);
    testY("1 (vmark memo) 1","1");
    starttest("cache hit, dyad NOT called");
    testassert(!dyadcalled);
    testY("(([: >: vmark) memo) 1","2");
    testY("1 (([: >: vmark) memo) 1","2");
    monadcalled=false;
    dyadcalled=false;
    testY("(([: >: vmark) memo) 1","2");
    starttest("cache hit, monad NOT called");
    testassert(!monadcalled);
    testY("1 (([: >: vmark) memo) 1","2");
    starttest("cache hit, dyad NOT called");
    testassert(!dyadcalled);
  }
  
  section("reporting");
  starttest("Checking that tests were properly terminated");
  testassert((tests_run-1)==(tests_ok+tests_fail));
  cout << "=== ran "<<tests_run<<" tests with "<<tests_ok<<" successes and "<<
    tests_fail<<" failures."<<endl;
  if (tests_ok==tests_run)
    cout << "ALL OK."<<endl;
  else
    cout << "TESTS FAILED !!" << endl;
  return tests_fail;
}

int main (int argc, char **argv) {
  return runTests();
} 
