/* -*- mode: c++; indent-tabs-mode:nil -*- */
#include <cstdlib>
#include <cstdio>
#include <cerrno>
#include <climits>
#include <limits>
#include "util.h"

int parseint(char *str, int& res) {
  char *end;
  errno = 0;
  const long sl = strtol(str, &end, 10);
  if (end == str) {
    fprintf(stderr, "%s: not a decimal number\n", str);
  }
  else if ('\0' != *end) {
    fprintf(stderr, "%s: extra characters at end: %s\n", str, end);
  }
  else if ((LONG_MIN == sl || LONG_MAX == sl) && ERANGE == errno) {
    fprintf(stderr, "%s out of range of type long\n", str);
  }
  else if (sl > INT_MAX) {
    fprintf(stderr, "%ld greater than INT_MAX\n", sl);
  }
  else if (sl < INT_MIN) {
    fprintf(stderr, "%ld less than INT_MIN\n", sl);
  } else {
    res=(int)sl;
    return 0;
  }
  return 1;
}

int parsepositiveint(char *str, int& res) {
  if ((parseint(str, res) == 0) && ( res >= 0 ))
    return 0;
  else
    return 1;
}

int parsedouble(char *str, double& res) {
  char *end;
  errno = 0;
  const double sd = strtol(str, &end, 10);
  if (end == str) {
    fprintf(stderr, "%s: not a floating point number\n", str);
  }
  else if ('\0' != *end) {
    fprintf(stderr, "%s: extra characters at end: %s\n", str, end);
  } else {
    res=sd;
    return 0;
  }
  return 1;
}

int parsepositivedouble(char *str, double& res) {
  if ((parsedouble(str, res) == 0) && ( res >= 0 ))
    return 0;
  else
    return 1;
}

std::string sha1tostring(SHA1 &sha1) {
  using namespace std;
  unsigned digest[5];
  sha1.Result(digest);
  ostringstream str;
  str.setf(ios::hex,ios::basefield);
  for(int i=0;i<5;i++)
    str << digest[i];
  return str.str();
};

const double EPS =  std::numeric_limits<double>::epsilon();

bool tol_eq(double a, double b) {                                            
  return fabs(a-b)<=EPS*1000*std::max(fabs(a),fabs(b));                       
};
